from langmuir import *
import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import least_squares
from tqdm import tqdm

geometry = Cylinder(r=0.255e-3, l=25e-3, lguard=True)
V = np.array([2,3,4,5])

model_truth = OML_current
model_pred = OML_current

data = generate_synthetic_data(geometry, V, model=model_truth)

alt = data['alt']
ne_jacobsen = jacobsen_density(geometry, V, data['I'])
ne_fit = np.zeros_like(alt)
V0_fit = np.zeros_like(alt)

def residual(x, I, T):
    n, V0 = x
    return (model_pred(geometry, Electron(n=n, T=T), V+V0) - I)*1e6

x0 = [10e10, -0.3]
x_scale = [1e10, 1]
for i in tqdm(range(len(alt))):
    I = data['I'][i]
    T = data['Te'][i]
    T = 2000
    res = least_squares(residual, x0, args=(I,T), x_scale=x_scale)
    ne_fit[i], V0_fit[i] = res.x
    x0 = res.x

plot = plt.plot

plt.figure()
plot(data['ne'], alt, label='Ground truth')
plot(ne_jacobsen, alt, label='Jacobsen')
plot(ne_fit, alt, label='Fit')
plt.xlabel('Density $[\mathrm{m}^{-3}]$')
plt.ylabel('Altitude $[\mathrm{km}]$')
plt.legend()

plt.figure()
plot(data['V0'], alt, label='Ground truth')
plot(V0_fit, alt, label='Fit')
plt.xlabel('Floating potential $[\mathrm{V}]$')
plt.ylabel('Altitude $[\mathrm{km}]$')
plt.legend()

plt.show()



from langmuir import *
from scipy.optimize import leastsq

geo = Cylinder(1e-3, 25e-3)
V0 = -0.5
V = np.array([2.5, 4.0, 5.5, 7.0])
I = OML_current(geo, Species(n=120e10, T=1000), V+V0)

def residual(x):
    n, V0 = x
    return OML_current(geo, Species(n=n, T=1500), V+V0) - I

x0 = [10e10, -0.3]
x, c = leastsq(residual, x0)
n, V0 = x
print(n)
print(V0)



"""
Copyright 2018
    Sigvald Marholm <marholm@marebakken.com>
    Diako Darian <diakod@math.uio.no>

This file is part of langmuir.

langmuir is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

langmuir is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with langmuir.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import division
from langmuir.species import *
from langmuir.geometry import *
from langmuir.misc import *
from scipy.constants import value as constants
from scipy.special import gamma, erfc, hyp2f1
from copy import deepcopy
import numpy as np

def normalization_current(geometry, species):
    """
    Returns the normalization current for the given species and geometry.
    The normalization current is the current the species would have contributed
    to a probe at zero potential with respect to the background plasma due to
    random thermal movements of particles, if the species had been Maxwellian.

    Parameters
    ----------
    geometry: Plane, Cylinder or Sphere
        Probe geometry

    species: Species or array-like of Species
        Species constituting the background plasma

    Returns
    -------
    float
    """

    if isinstance(species, list):
        I = 0
        for s in species:
            I += normalization_current(geometry, s)
        return I

    q, n, vth = species.q, species.n, species.vth

    if isinstance(geometry, Sphere):
        r = geometry.r
        I0 = 2*np.sqrt(2*np.pi)*r**2*q*n*vth

    elif isinstance(geometry, Cylinder):
        r, l = geometry.r, geometry.l
        I0 = np.sqrt(2*np.pi)*r*l*q*n*vth

    else:
        raise ValueError('Geometry not supported: {}'.format(geometry))

    return I0

def thermal_current(geometry, species, normalization=None):
    """
    Returns the thermal current.

    Parameters
    ----------
    geometry: Plane, Cylinder or Sphere
        Probe geometry

    species: Species or array-like of Species
        Species constituting the background plasma

    Returns
    -------
    float
    """

    if isinstance(species, list):
        if normalization is not None:
            logger.error('Cannot normalize current to more than one species')
            return None
        I = 0
        for s in species:
            I += thermal_current(geometry, s)
        return I

    kappa, alpha = species.kappa, species.alpha

    if normalization is None:
        I0 = normalization_current(geometry, species)
    elif normalization.lower() == 'thmax':
        I0 = 1
    else:
        raise ValueError('Normalization not supported: {}'.format(normalization))

    if isinstance(geometry, Sphere):
        if kappa == float('inf'):
            C = 1.0
            D = (1.+24*alpha)/(1.+15*alpha)
        else:
            C = np.sqrt(kappa-1.5)*gamma(kappa-1.)/gamma(kappa-0.5)
            D = (1.+24*alpha*((kappa-1.5)**2/((kappa-2.)*(kappa-3.))))/(1.+15*alpha*((kappa-1.5)/(kappa-2.5)))
    elif isinstance(geometry, Cylinder):
        if kappa == float('inf'):
            C = 1.0
            D = (1.+3.*alpha)/(1.+15.*alpha)
        else:
            C = np.sqrt(kappa - 1.5) * (kappa - .5) / (kappa - 1.0)
            D = (1. + 3 * alpha * ((kappa - 1.5) / (kappa - 0.5))) / \
                (1. + 15 * alpha * ((kappa - 1.5) / (kappa - 2.5)))
    else:
        raise ValueError('Geometry not supported: {}'.format(geometry))

    return I0*C*D

def OML_current(geometry, species, V=None, eta=None, normalization=None):
    """
    Current collected by a probe according to the Orbital Motion Limited (OML)
    theory. The model assumes a probe of infinitely small radius compared to
    the Debye length, and for a cylindrical probe, that it is infinitely long.
    Probes with radii up to 0.2 Debye lengths (for spherical probes) or 1.0
    Debye lengths (for cylindrical probes) are very well approximated by this
    theory, although the literature is diverse as to how long cylindrical probes
    must be for this theory to be a good approximation.

    Parameters
    ----------
    geometry: Plane, Cylinder or Sphere
        Probe geometry

    species: Species or array-like of Species
        Species constituting the background plasma

    V: float or array-like of floats
        Probe voltage(s) in [V]. Overrides eta.

    eta: float or array-like of floats
        Probe voltage(s) normalized by k*T/q, where q and T are the species'
        charge and temperature and k is Boltzmann's constant.

    normalization: 'th', 'thmax', None
        Wether to normalize the output current by, respectively, the thermal
        current, the Maxwellian thermal current, or not at all, i.e., current
        in [A/m].

    Returns
    -------
    float if voltage is float. array of floats corresponding to voltage if
    voltage is array-like.
    """

    if isinstance(species, list):
        if normalization is not None:
            logger.error('Cannot normalize current to more than one species')
            return None
        if eta is not None:
            logger.error('Cannot normalize voltage to more than one species')
            return None
        I = 0
        for s in species:
            I += OML_current(geometry, s, V, eta)
        return I

    q, T = species.q, species.T
    kappa, alpha = species.kappa, species.alpha

    k = constants('Boltzmann constant')

    if V is not None:
        eta_isarray = isinstance(V, (np.ndarray, list, tuple))
        V = make_array(V)
        eta = -q*V/(k*T)
    else:
        eta_isarray = isinstance(eta, (np.ndarray, list, tuple))
        eta = make_array(eta)

    eta = deepcopy(eta) # Prevents this function from changing eta in caller

    I = np.zeros_like(eta, dtype=float)

    eta[np.where(eta==0)] = np.finfo(float).eps # replace zeros with machine eps
    indices_n = np.where(eta < 0)[0]   # indices for repelled particles
    indices_p = np.where(eta >= 0)[0]  # indices for attracted particles

    if kappa == float('inf'):
        C = 1.0
        D = (1.+24*alpha)/(1.+15*alpha)
        E = 4.*alpha/(1.+24*alpha)
        F = (1.+8*alpha)/(1.+24*alpha)
    else:
        C = np.sqrt(kappa-1.5)*gamma(kappa-1.)/gamma(kappa-0.5)
        D = (1.+24*alpha*((kappa-1.5)**2/((kappa-2.)*(kappa-3.))))/(1.+15*alpha*((kappa-1.5)/(kappa-2.5)))
        E = 4.*alpha*kappa*(kappa-1.)/( (kappa-2.)*(kappa-3.)+24*alpha*(kappa-1.5)**2 )
        F = ((kappa-1.)*(kappa-2.)*(kappa-3.)+8*alpha*(kappa-3.)*(kappa-1.5)**2) /( (kappa-2.)*(kappa-3.)*(kappa-1.5)+24*alpha*(kappa-1.5)**3 )

    if normalization is None:
        I0 = normalization_current(geometry, species)
    elif normalization.lower() == 'thmax':
        I0 = 1
    elif normalization.lower() == 'th':
        I0 = normalization_current(geometry, species)/\
             thermal_current(geometry, species)
    else:
        raise ValueError('Normalization not supported: {}'.format(normalization))

    if isinstance(geometry, Sphere):

        # repelled particles:
        if species.kappa == float('inf'):
            I[indices_n] = I0 * C * D * np.exp(eta[indices_n]) * (1. + E * -eta[indices_n] * (-eta[indices_n] + 4.))

        else:
            I[indices_n] = I0 * C * D * (1. - eta[indices_n] / (kappa - 1.5))**(
                1. - kappa) * (1. + E * (-eta[indices_n]) * (-eta[indices_n] + 4. * ((kappa - 1.5) / (kappa - 1.))))

        # attracted particles:
        eta[indices_p] = np.abs(eta[indices_p])
        I[indices_p] = I0*C*D*(1.+F*eta[indices_p])

    elif isinstance(geometry, Cylinder):

        # repelled particles:
        if species.kappa == float('inf'):
            I[indices_n] = I0 * C * D * \
                np.exp(eta[indices_n]) * (1. + E *
                                           (-eta[indices_n]) * (-eta[indices_n] + 4.))

        else:
            I[indices_n] = I0 * C * D * (1. - eta[indices_n] / (kappa - 1.5))**(
                1. - kappa) * (1. + E * (-eta[indices_n]) * (-eta[indices_n] + 4. * ((kappa - 1.5) / (kappa - 1.))))

        # attracted particles:
        eta[indices_p] = np.abs(eta[indices_p])
        if species.kappa == float('inf'):
            I[indices_p] = I0 * C * D * \
                           ((2. / np.sqrt(np.pi)) * ( 1 - 0.5 * E * eta[indices_p]) * np.sqrt(eta[indices_p]) + \
                           np.exp(eta[indices_p]) * (1. + E * eta[indices_p] * (eta[indices_p] - 4.)) * erfc(np.sqrt(eta[indices_p])))

        else:
            C = np.sqrt(kappa - 1.5) * (kappa - .5) / (kappa - 1.0)
            D = (1. + 3 * alpha * ((kappa - 1.5) / (kappa - 0.5))) / \
                (1. + 15 * alpha * ((kappa - 1.5) / (kappa - 2.5)))
            E = 4. * alpha * kappa * \
                (kappa - 1.) / ((kappa - .5) *
                                (kappa - 1.5) + 3. * alpha * (kappa - 1.5)**2)

            I[indices_p] = (2./np.sqrt(np.pi))*I0 * C * D * (eta[indices_p]/(kappa-1.5))**(1.-kappa) * \
                (((kappa - 1.) / (kappa - 3.)) * E * (eta[indices_p]**2) * hyp2f1(kappa - 3, kappa + .5, kappa - 2., 1. - (kappa - 1.5) / (eta[indices_p])) + \
                ((kappa - 1.5 - 2. * (kappa - 1.) * eta[indices_p]) / (kappa - 2.)) * E * eta[indices_p] * hyp2f1(kappa - 2, kappa + .5, kappa - 1., 1. - (kappa - 1.5) / (eta[indices_p])) +
                (1. + E * eta[indices_p] * (eta[indices_p]-((kappa-1.5)/(kappa-1.)))) * hyp2f1(kappa - 1., kappa + .5, kappa, 1. - (kappa - 1.5) / (eta[indices_p])))

    else:
        raise ValueError('Geometry not supported: {}'.format(geometry))

    return I if eta_isarray else I[0]

def jacobsen_density(geometry, biases, currents, species=Electron()):
    """
    Density computed according to the slope in current squared versus voltage
    (the Jacobsen-Bekkeng method). This assumes that OML theory for a
    cylindrical probe is exact.

    Parameters
    ----------
    geometry: Cylinder
        Probe geometry

    biases: array-like of floats
        Probe bias voltages [V] with respect to some common reference

    currents: 2D array-like of floats
        Current measurements. currents[i,j] is sample i, probe current j

    species: Species
        The attracted species. The density and temperature in the species
        object is disregarded by this function.

    Returns
    -------
    Array of computed densities, one element for each row in currents.
    """

    currents = make_array(currents)
    biases = make_array(biases)

    if not isinstance(geometry, Cylinder):
        raise ValueError('Geometry not supported: {}'.format(geometry))

    if len(biases) != currents.shape[1]:
        raise ValueError('The number of columns in currents must equal the'\
                         'number of elements in biases')

    m = species.m
    q = species.q
    k = constants('Boltzmann constant')
    C = 2/np.sqrt(np.pi)
    S = 2*np.pi*geometry.r*geometry.l # surface area
    slope_factor = -2*np.pi*m/((C*S)**2*q**3)

    m = len(biases)
    sv = np.sum(biases)
    svv = np.sum(biases**2)
    curr_sq = currents**2
    si = np.sum(curr_sq, axis=1)
    svi = np.sum(biases*curr_sq, axis=1)
    slope = (m*svi-sv*si)/(m*svv-sv**2)
    density = np.sqrt(slope_factor*slope)

    return density



from langmuir import *
import numpy as np
import matplotlib.pyplot as plt

elec = Electron(n=1e11, T=1000)

l_probe = 40e-3
l_guard = 25e-3
z_probe = np.linspace(0, l_probe, 100)
z_guard = np.linspace(-l_guard, 0, 100)

geo = Cylinder(r=0.25e-3, l=l_probe, lguard=True)
i_probe = finite_length_current_density(geo, elec, V=5, z=z_probe)
i_guard = finite_length_current_density(geo, elec, V=5, z=z_guard)

plt.plot(z_probe*1e3, -i_probe*1e6)
plt.plot(z_guard*1e3, -i_guard*1e6, ':C0')
plt.xlabel('z [mm]')
plt.ylabel('i [nA/mm]')
plt.show()



from scipy.constants import value as constants
import matplotlib.pyplot as plt
from datetime import datetime
import numpy as np
import pyglow
from tqdm import tqdm

lat = 45.
lon = 0.
alts = np.linspace(60., 1300., 1001)
dn = datetime(2008, 3, 1, 12, 00)

ns = {}
Ts = {}

for alt in tqdm(alts):
    pt = pyglow.Point(dn, lat, lon, alt)
    pt.run_iri()
    pt.run_msis()

    n = pt.ni
    n['e'] = pt.ne
    for key in n:
        value = n[key]*1e6
        if value < 1e8: value = float('nan')
        if not key in ns:
            ns[key] = []
        ns[key].append(value)

    T = {'e': pt.Te, 'i': pt.Ti}
    for key in T:
        value = T[key]
        if value < 1: value = float('nan')
        if not key in Ts:
            Ts[key] = []
        Ts[key].append(value)

np.savez_compressed('iri.npz',
                    alt=alts,
                    Te=np.array(Ts['e']),
                    Ti=np.array(Ts['i']),
                    ne=np.array(ns['e']),
                    nO=np.array(ns['O+']),
                    nO2=np.array(ns['O2+']),
                    nNO=np.array(ns['NO+']),
                    nH=np.array(ns['H+'])
                   )


plt.figure()
plot = plt.loglog
plot(ns['e'], alts, label='e')
plot(ns['O+'], alts, label='O+')
plot(ns['O2+'], alts, label='O2+')
plot(ns['NO+'], alts, label='NO+')
plot(ns['H+'], alts, label='H+')
plt.legend()

plt.figure()
plot = plt.semilogy
plot(Ts['e'], alts, label='e')
plot(Ts['i'], alts, label='i')
plt.legend()

plt.show()



"""
Copyright 2018
    Sigvald Marholm <marholm@marebakken.com>
    Diako Darian <diakod@math.uio.no>

This file is part of langmuir.

langmuir is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

langmuir is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with langmuir.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import division
from langmuir import *

def test_make_array():
    assert(isinstance(make_array(2), np.ndarray))
    assert(isinstance(make_array((2,3)), np.ndarray))
    assert(isinstance(make_array([2,3]), np.ndarray))
    assert(np.allclose(2, make_array(  int(2))))
    assert(np.allclose(2, make_array(float(2))))
    assert(np.allclose([10, 1/10], make_array([10, 1/10])))



from langmuir import *
import numpy as np
import matplotlib.pyplot as plt

plasma = [
    Electron(n=1e11, T=1000),
    Hydrogen(n=1e11, T=1000)
    ]

geometry = Cylinder(r=1e-3, l=60e-3, lguard=True)

V = np.linspace(-2, 2, 100)

I_OML = OML_current(geometry, plasma, V)
I_FL = finite_length_current(geometry, plasma, V)

plt.plot(V, -I_OML*1e6, label='OML')
plt.plot(V, -I_FL*1e6, label='FL')
plt.xlabel('V [V]')
plt.ylabel('I [uA]')
plt.legend()
plt.show()



from langmuir import *
from scipy.constants import value as constants
from scipy.optimize import root_scalar

n=1e11
T=1000
e = constants('elementary charge')
k = constants('Boltzmann constant')

plasma = [Electron(n=n, T=T),
          Hydrogen(n=n, T=T)]

geometry = Sphere(r=0.2*debye(plasma))

def res(V):
    return OML_current(geometry, plasma, V)

sol = root_scalar(res, x0=-3, x1=0)

print(sol.root)
print(-2.5*k*T/e)



from langmuir import *
import numpy as np
import matplotlib.pyplot as plt

plasma = Electron(n=1e11, T=1000)
geometry = Sphere(r=5*debye(plasma))

V = np.linspace(0, 2, 100)

I_OML = OML_current(geometry, plasma, V)
I_FR = finite_radius_current(geometry, plasma, V)

plt.plot(V, -I_OML*1e6, label='OML')
plt.plot(V, -I_FR*1e6, label='FR')
plt.xlabel('V [V]')
plt.ylabel('I [uA]')
plt.legend()
plt.show()



from langmuir import *
import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit

def power_law(eta, c, beta):
    return -c*eta**beta

elec = Species(n=1e11, T=1000)
geo = Cylinder(l=40e-3, r=1e-3, lguard=0.77)
geo = Cylinder(l=20e-3, r=1e-3, lguard=0.077)
geo = Cylinder(l=40e-3, r=1e-3, lguard=0.077)
# geo = Cylinder(l=5e-3, r=1e-3, lguard=100e-3)
geo = Cylinder(l=10*elec.debye, r=1e-3, lguard=200*elec.debye)
# geo = Cylinder(l=600.0*elec.debye, r=1e-3, lguard=0.0*elec.debye)
etas = np.linspace(-10,100,400)

# Is = finite_length_current(geo, elec, eta=etas)
# plt.plot(etas, Is)
# plt.show()

Vs = np.linspace(0, 8, 100)
Is_I = -finite_length_current(geo, elec, eta=-etas)
# Is_I2 = -finite_length_current(geo, elec, eta=-etas, interpolate='i2')
Is_g = -finite_length_current(geo, elec, eta=-etas, interpolate='g')

popt_g, pcov_g = curve_fit(power_law, etas, Is_g)
Is_g_fit = power_law(etas, *popt_g)
print("g beta:", popt_g[1])

popt_I, pcov_I = curve_fit(power_law, etas, Is_I)
Is_I_fit = power_law(etas, *popt_I)
print("I beta:", popt_I[1])

# popt_I2, pcov_I2 = curve_fit(power_law, etas, Is_I2)
# Is_I2_fit = power_law(etas, *popt_I2)
# print("I2 beta:", popt_I2[1])

plt.plot(etas, Is_I, label='I')
# plt.plot(etas, Is_I2, label='I2')
plt.plot(etas, Is_g, label='g')
plt.plot(etas, Is_g_fit, label='g_fit')
plt.plot(etas, Is_I_fit, label='I_fit')
# plt.plot(etas, Is_I2_fit, label='I2_fit')
plt.legend()
plt.show()

# TODO:
# Potentials <0 and >100



from langmuir import *
import numpy as np
import matplotlib.pyplot as plt

V = np.linspace(-2, 2, 100)
I = OML_current(Cylinder(r=1e-3, l=25e-3), Electron(n=1e11, T=1000), V)

plt.plot(V, -I*1e6)
plt.xlabel('V [V]')
plt.ylabel('I [uA]')
plt.show()



"""
Copyright 2018
    Sigvald Marholm <marholm@marebakken.com>
    Diako Darian <diakod@math.uio.no>

This file is part of langmuir.

langmuir is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

langmuir is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with langmuir.  If not, see <http://www.gnu.org/licenses/>.
"""

from scipy.constants import value as constants
from scipy.constants import epsilon_0
from langmuir.misc import *
import numpy as np

class Species(object):
    """
    Defines a species using a set of flags and keyword parameters. Maxwellian
    electrons are the default but Kappa, Cairns, and Kappa-Cairns distributions
    can be specified by setting the spectral indices kappa and alpha. Other
    parameters can be specified/overridden using the keyword parameters.
    Examples::

        >>> # Maxwellian electrons with density 1e11 m^(-3) and 1000 K
        >>> species = Electron(n=1e11, T=1000)

        >>> # Kappa-distributed electrons with kappa=2
        >>> species = Electron(n=1e11, T=1000, kappa=2)

        >>> # Doubly charged Maxwellain Oxygen ions with 0.26 eV
        >>> species = Species(Z=2, amu=16, n=1e11, eV=0.26)

        >>> # Kappa-Cairns-distributed protons with kappa = 5 and alpha=0.2
        >>> species = Proton(n=1e11, T=1000, kappa=5, alpha=0.2)

    A plasma is fully specified by a list of species. E.g. for a Maxwellian
    electron-proton plasma::

        >>> plasma = []
        >>> plasma.append(Electron(n=1e11, T=1000))
        >>> plasma.append(Proton(n=1e11, T=1000))

    Keyword parameters:
    -------------------
    m     : mass [kg]
    amu   : mass [amu]
    q     : charge [C]
    Z     : charge [elementary charges]
    n     : density [m^(-3)]
    T     : temperature [K]
    eV    : temperature [eV]
    vth   : thermal velocity [m/s]
    kappa : spectral index kappa for Kappa and Kappa-Cairns distribution
    alpha : spectral index alpha for Cairns and Kappa-Cairns distribution
    """

    def_Z   = 1
    def_amu = constants('electron mass')/constants('atomic mass constant')

    def __init__(self, **kwargs):

        e    = constants('elementary charge')
        k    = constants('Boltzmann constant')
        eps0 = epsilon_0 #constants('vacuum electric permittivity')
        amu  = constants('atomic mass constant')

        if 'Z' in kwargs:
            self.q = kwargs['Z']*e
        elif 'q' in kwargs:
            self.q = kwargs['q']
        else:
            self.q = self.def_Z*e

        if 'amu' in kwargs:
            self.m = kwargs['amu']*amu
        elif 'm' in kwargs:
            self.m = kwargs['m']
        else:
            self.m = self.def_amu*amu

        if 'eV' in kwargs:
            self.T = kwargs['eV']*e/k
        elif 'vth' in kwargs:
            self.T = self.m*kwargs['vth']**2/k
        elif 'T' in kwargs:
            self.T = kwargs['T']
        else:
            self.T = 1000

        if self.T<0:
            self.T=0
            logger.warning('Negative temperature interpreted as zero')

        self.n     = kwargs.pop('n', 1e11)
        self.alpha = kwargs.pop('alpha', 0)
        self.kappa = kwargs.pop('kappa', float('inf'))

        self.Z = self.q/e
        self.amu = self.m/amu
        self.eV  = self.T*k/e
        self.vth = np.sqrt(k*self.T/self.m)
        self.omega_p = np.sqrt(self.q**2*self.n/(eps0*self.m))
        self.freq_p = self.omega_p/(2*np.pi)
        self.period_p = 1/self.freq_p

        if self.kappa == float('inf'):
            self.debye = np.sqrt(eps0*k*self.T/(self.q**2*self.n))*np.sqrt((1.0 + 15.0*self.alpha)/(1.0 + 3.0*self.alpha))
        else:
            self.debye = np.sqrt(eps0 * k * self.T / (self.q**2 * self.n)) *\
                         np.sqrt(((self.kappa - 1.5) / (self.kappa - 0.5)) *\
                        ((1.0 + 15.0 * self.alpha * ((self.kappa - 1.5) / (self.kappa - 2.5))) / (1.0 + 3.0 * self.alpha * ((self.kappa - 1.5) / (self.kappa - 0.5)))))

    def omega_c(self, B):
        """
        Returns the angular cyclotron frequency of the species for a given B-field
        """
        return np.abs(B*self.q/self.m)

    def freq_c(self, B):
        """
        Returns the cyclotron frequency of the species for a given B-field
        """
        return self.omega_c(B)/(2*np.pi)

    def period_c(self, B):
        """
        Returns the cyclotron period of the species for a given B-field
        """
        return 1/self.freq_c(B)

    def larmor(self, B):
        """
        Returns the Larmor radius of the species for a given B-field
        """
        return self.vth/self.omega_c(B)

    def __repr__(self):

        s = "Species(q={}, m={}, n={}, T={}".format(self.q, self.m, self.n, self.T)

        if(self.kappa != float('inf')):
            s += ", kappa={}".format(self.kappa)

        if(self.alpha != 0):
            s += ", alpha={}".format(self.alpha)

        s += ")"

        return s

def debye(species):
    """
    Get the total Debye length of a species or a plasma (represented as a list
    of species). If species is a list this includes the Debye length due to
    all species.
    """
    if not isinstance(species, list):
        species = [species]
    return sum([s.debye**(-2) for s in species])**(-0.5)

class Electron(Species): def_Z = -1
class Proton(Species): def_amu = 1
class Positron(Species): pass
class Antiproton(Species): def_Z = -1; def_amu = 1
class Vogon(Species): def_Z = 1j
class Hydrogen(Species): def_amu = 1.0080
class Helium(Species): def_amu = 4.00260
class Lithium(Species): def_amu = 7.0
class Beryllium(Species): def_amu = 9.012183
class Boron(Species): def_amu = 10.81
class Carbon(Species): def_amu = 12.011
class Nitrogen(Species): def_amu = 14.007
class Oxygen(Species): def_amu = 15.999
class Fluorine(Species): def_amu = 18.99840316
class Neon(Species): def_amu = 20.180
class Sodium(Species): def_amu = 22.9897693
class Magnesium(Species): def_amu = 24.305
class Aluminum(Species): def_amu = 26.981538
class Silicon(Species): def_amu = 28.085
class Phosphorus(Species): def_amu = 30.97376200
class Sulfur(Species): def_amu = 32.07
class Chlorine(Species): def_amu = 35.45
class Argon(Species): def_amu = 39.9
class Potassium(Species): def_amu = 39.098
class Calcium(Species): def_amu = 40.08
class Scandium(Species): def_amu = 44.95591
class Titanium(Species): def_amu = 47.87
class Vanadium(Species): def_amu = 50.941
class Chromium(Species): def_amu = 51.996
class Manganese(Species): def_amu = 54.93804
class Iron(Species): def_amu = 55.84
class Cobalt(Species): def_amu = 58.93319
class Nickel(Species): def_amu = 58.693
class Copper(Species): def_amu = 63.55
class Zinc(Species): def_amu = 65.4
class Gallium(Species): def_amu = 69.72
class Germanium(Species): def_amu = 72.63
class Arsenic(Species): def_amu = 74.92159
class Selenium(Species): def_amu = 78.97
class Bromine(Species): def_amu = 79.90
class Krypton(Species): def_amu = 83.80
class Rubidium(Species): def_amu = 85.468
class Strontium(Species): def_amu = 87.6
class Yttrium(Species): def_amu = 88.9058
class Zirconium(Species): def_amu = 91.22
class Niobium(Species): def_amu = 92.9064
class Molybdenum(Species): def_amu = 96.0
class Technetium(Species): def_amu = 97.90721
class Ruthenium(Species): def_amu = 101.1
class Rhodium(Species): def_amu = 102.9055
class Palladium(Species): def_amu = 106.4
class Silver(Species): def_amu = 107.868
class Cadmium(Species): def_amu = 112.41
class Indium(Species): def_amu = 114.82
class Tin(Species): def_amu = 118.71
class Antimony(Species): def_amu = 121.76
class Tellurium(Species): def_amu = 127.6
class Iodine(Species): def_amu = 126.9045
class Xenon(Species): def_amu = 131.29
class Cesium(Species): def_amu = 132.9054520
class Barium(Species): def_amu = 137.33
class Lanthanum(Species): def_amu = 138.9055
class Cerium(Species): def_amu = 140.12
class Praseodymium(Species): def_amu = 140.9077
class Neodymium(Species): def_amu = 144.24
class Promethium(Species): def_amu = 144.91276
class Samarium(Species): def_amu = 150.4
class Europium(Species): def_amu = 151.96
class Gadolinium(Species): def_amu = 157.2
class Terbium(Species): def_amu = 158.92535
class Dysprosium(Species): def_amu = 162.50
class Holmium(Species): def_amu = 164.93033
class Erbium(Species): def_amu = 167.26
class Thulium(Species): def_amu = 168.93422
class Ytterbium(Species): def_amu = 173.04
class Lutetium(Species): def_amu = 174.967
class Hafnium(Species): def_amu = 178.5
class Tantalum(Species): def_amu = 180.9479
class Tungsten(Species): def_amu = 183.8
class Rhenium(Species): def_amu = 186.21
class Osmium(Species): def_amu = 190.2
class Iridium(Species): def_amu = 192.22
class Platinum(Species): def_amu = 195.08
class Gold(Species): def_amu = 196.96657
class Mercury(Species): def_amu = 200.59
class Thallium(Species): def_amu = 204.383
class Lead(Species): def_amu = 207
class Bismuth(Species): def_amu = 208.9804
class Polonium(Species): def_amu = 208.98243
class Astatine(Species): def_amu = 209.98715
class Radon(Species): def_amu = 222.01758
class Francium(Species): def_amu = 223.01973
class Radium(Species): def_amu = 226.02541
class Actinium(Species): def_amu = 227.02775
class Thorium(Species): def_amu = 232.038
class Protactinium(Species): def_amu = 231.0359
class Uranium(Species): def_amu = 238.0289
class Neptunium(Species): def_amu = 237.04817
class Plutonium(Species): def_amu = 244.06420
class Americium(Species): def_amu = 243.06138
class Curium(Species): def_amu = 247.07035
class Berkelium(Species): def_amu = 247.07031
class Californium(Species): def_amu = 251.07959
class Einsteinium(Species): def_amu = 252.0830
class Fermium(Species): def_amu = 257.09511
class Mendelevium(Species): def_amu = 258.09843
class Nobelium(Species): def_amu = 259.10100
class Lawrencium(Species): def_amu = 262.110
class Rutherfordium(Species): def_amu = 267.122
class Dubnium(Species): def_amu = 268.126
class Seaborgium(Species): def_amu = 271.134
class Bohrium(Species): def_amu = 274.144
class Hassium(Species): def_amu = 277.152
class Meitnerium(Species): def_amu = 278.156
class Darmstadtium(Species): def_amu = 281.165
class Roentgenium(Species): def_amu = 282.169
class Copernicium(Species): def_amu = 285.177
class Nihonium(Species): def_amu = 286.183
class Flerovium(Species): def_amu = 289.191
class Moscovium(Species): def_amu = 290.196
class Livermorium(Species): def_amu = 293.205
class Tennessine(Species): def_amu = 294.211
class Oganesson(Species): def_amu = 294.214
# Extracted from PubChem



from langmuir import *
import numpy as np
import matplotlib.pyplot as plt

elec = Electron(n=1e11, T=1000)
l = 40e-3
z = np.linspace(0,l,100)

geo = Cylinder(r=0.25e-3, l=l)
i = finite_length_current_density(geo, elec, V=5, z=z)

plt.plot(z*1e3, -i*1e6)
plt.xlabel('z [mm]')
plt.ylabel('i [nA/mm]')
plt.show()



from langmuir import *
import numpy as np
import matplotlib.pyplot as plt

elec = Electron()

lambds = np.array([10, 30, 100])
eta = np.linspace(0, 100, 100)

for lambd in lambds:
    geometry = Cylinder(r=0.2*elec.debye, l=lambd*elec.debye)
    current = finite_length_current(geometry, elec, eta=eta, normalization='th')
    plt.plot(eta, current, label='$\lambda={}$'.format(lambd))

current = OML_current(geometry, elec, eta=eta, normalization='th')
plt.plot(eta, current, label='OML')

plt.xlabel('$\eta$')
plt.ylabel('$I/I_\mathrm{th}$')
plt.legend()
plt.show()



"""
Copyright 2018
    Sigvald Marholm <marholm@marebakken.com>
    Diako Darian <diakod@math.uio.no>

This file is part of langmuir.

langmuir is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

langmuir is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with langmuir.  If not, see <http://www.gnu.org/licenses/>.
"""
from langmuir.finite_length import *
from scipy.optimize import root_scalar

def generate_synthetic_data(geometry,
                            V,
                            model=finite_length_current,
                            V0=None,
                            alt_range=(100,500),
                            noise=1e-5,
                           ):

    fname = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        'iri.npz')

    iri_data = np.load(fname)

    res = {}
    ind, = np.where(np.logical_and(iri_data['alt']>=alt_range[0],
                                   iri_data['alt']<=alt_range[1]))
    res['alt'] = iri_data['alt'][ind]
    res['Te'] = iri_data['Te'][ind]
    res['Ti'] = iri_data['Ti'][ind]
    res['ne'] = iri_data['ne'][ind]
    res['nO+'] = iri_data['nO'][ind]
    res['nO2+'] = iri_data['nO2'][ind]
    res['nNO+'] = iri_data['nNO'][ind]
    res['nH+'] = iri_data['nH'][ind]

    V = make_array(V)

    I = np.zeros((len(res['alt']), len(V)))
    V0_ = np.ones(len(res['alt']))
    if V0 is not None:
        V0_ *= V0

    def cap(x):
        if np.isnan(x):
            return 1 # FIXME: Finite-length model does not support zero density
        else:
            return x

    for i in range(len(res['alt'])):

        Te = res['Te'][i]
        Ti = res['Ti'][i]

        plasma = [Electron(n=cap(res['ne'][i]), T=Te)
                 ,Oxygen(n=cap(res['nO+'][i]), T=Ti)
                 ,Oxygen(n=cap(res['nO2+'][i]), T=Ti, Z=2)
                 ,Species(n=cap(res['nNO+'][i]), T=Ti, amu=30)
                 ,Hydrogen(n=cap(res['nH+'][i]), T=Ti)
                 ]

        if V0 is None:
            # This is just to have some varying number for the floating potential.
            # It doesn't have to be realistic, just within a resonable value
            residual_geometry = Sphere(r=10*debye(plasma))
            def residual(V):
                return OML_current(residual_geometry, plasma, V)
            sol = root_scalar(residual, x0=-10, x1=0)
            V0_[i] = sol.root

        try:
            I[i] = model(geometry, plasma, V=V0_[i]+V)
        except:
            I[i] = float('nan')

    I += noise*np.sqrt(np.abs(I))*np.random.randn(*I.shape)

    res['I'] = I
    res['V0'] = V0_

    return res




from langmuir import *
import matplotlib.pyplot as plt
import numpy as np
from scipy.optimize import curve_fit

elec = Electron()
eta = np.linspace(10,100,100)
geo = Cylinder(r=0.1*elec.debye, l=10*elec.debye)
I = finite_length_current(geo, elec, eta=eta, normalization='th')

def power_law(eta, c, beta):
    return c*eta**beta

popt, pcov = curve_fit(power_law, eta, I)

plt.plot(eta, I, label='Finite-length model')
plt.plot(eta, power_law(eta, *popt), ':k',
         label=r'Power law ($c={:.2f}, \beta={:.2f})$'.format(*popt))

plt.xlabel(r'$\eta$')
plt.ylabel(r'$I/I_\mathrm{th}$')
plt.legend()
plt.show()



from langmuir import *
import numpy as np
import matplotlib.pyplot as plt

elec = Electron()

eta = np.linspace(0, 100, 100)

geometry = Cylinder(r=0.2*elec.debye, l=10*elec.debye)
current = OML_current(geometry, elec, eta=eta, normalization='th')

plt.plot(eta, current)
plt.xlabel('$\eta$')
plt.ylabel('$I/I_\mathrm{th}$')
plt.show()
